@@ -9,6 +9,14 @@ use Cwd;
 use 5.006;
 select(($|=1,select(STDERR),$|=1)[1]);
 
+my $testfh;
+if (exists $ENV{TEST_OUTPUT}) {
+	my $file = $ENV{TEST_OUTPUT};
+	open $testfh, '>>', $file or die qq{Could not append file "$file": $!\n};
+	Test::More->builder->failure_output($testfh);
+	Test::More->builder->todo_output($testfh);
+}
+
 my @schemas =
 	(
 	 'dbd_pg_testschema',
@@ -25,6 +33,7 @@ my @tables =
 	 'dbd_pg_test2',
 	 'dbd_pg_test1',
 	 'dbd_pg_test',
+	 'dbd_pg_test_geom',
 	 );
 
 my @sequences =
@@ -45,7 +54,7 @@ use vars qw/$fh/;
 sub connect_database {
 
 	## Connect to the database (unless 'dbh' is passed in)
-	## Setup all the tables (unless 'nosetup' is passed in)
+	## Setup all the tables (unless 'nocreate' is passed in)
 	## Returns three values:
 	## 1. helpconnect for use by 01connect.t
 	## 2. Any error generated
@@ -62,7 +71,8 @@ sub connect_database {
 	## We'll try various ways to get to a database to test with
 
 	## First, check to see if we've been here before and left directions
-	my ($testdsn,$testuser,$helpconnect,$su,$uid,$testdir,$pg_ctl,$initdb,$error) = get_test_settings();
+	my ($testdsn,$testuser,$helpconnect,$su,$uid,$testdir,$pg_ctl,$initdb,$error,$version)
+		= get_test_settings();
 
 	## Did we fail last time? Fail this time too, but quicker!
 	if ($testdsn =~ /FAIL!/) {
@@ -82,16 +92,16 @@ sub connect_database {
 			$testdsn =~ s/$alias\s*=/$arg->{dbreplace}=/;
 		}
 		if ($arg->{dbquotes}) {
-			$testdsn =~ s/$alias\s*=(\w+)/'db="'.lc $2.'"'/e;
+			$testdsn =~ s/$alias\s*=([\-\w]+)/'db="'.lc $2.'"'/e;
 		}
 
 		goto GOTDBH if eval {
-			$dbh = DBI->connect($testdsn, $testuser, '',
+			$dbh = DBI->connect($testdsn, $testuser, $ENV{DBI_PASS},
 								{RaiseError => 1, PrintError => 0, AutoCommit => 1});
 			1;
 		};
 
-		if ($@ =~ /invalid connection option/) {
+		if ($@ =~ /invalid connection option/ or $@ =~ /failed:.*"dbbarf"/) {
 			return $helpconnect, $@, undef;
 		}
 
@@ -123,11 +133,16 @@ sub connect_database {
 						my $sockdir = "$testdir/data/socket";
 						if (! -e $sockdir) {
 							mkdir $sockdir;
-							if (! chown $uid, -1, $sockdir) {
-								warn "chown of $sockdir failed!\n";
+							if ($uid) {
+								if (! chown $uid, -1, $sockdir) {
+									warn "chown of $sockdir failed!\n";
+								}
 							}
 						}
 						$option = q{-o '-k socket'};
+						if ($version <= 8.0) {
+							$option = q{-o '-k dbdpg_test_database/data/socket'};
+						}
 					}
 					my $COM = qq{$pg_ctl $option -l $testdir/dbdpg_test.logfile -D $testdir/data start};
 					if ($su) {
@@ -229,7 +244,7 @@ sub connect_database {
 		}
 
 	  INITDB:
-		my ($info,$testport);
+		my $testport;
 		$helpconnect = 16;
 
 		## Use the initdb found by App::Info
@@ -239,12 +254,14 @@ sub connect_database {
 		}
 
 		## Make sure initdb exists and is working properly
+		$ENV{LANG} = 'C';
 		$info = '';
 		eval {
-			$info = qx{$initdb --help 2>&1};
+			$info = qx{$initdb --version 2>&1};
 		};
 		last GETHANDLE if $@; ## Fail - initdb bad
-		if (!defined $info or ($info !~ /\@postgresql\.org/ and $info !~ /run as root/)) {
+		$version = 0;
+		if (!defined $info or ($info !~ /(Postgres)/i and $info !~ /run as root/)) {
 			if (defined $info) {
 				if ($info !~ /\w/) {
 					$@ = 'initdb not found: cannot run full tests without a Postgres database';
@@ -261,6 +278,12 @@ sub connect_database {
 			}
 			last GETHANDLE; ## Fail - initdb bad
 		}
+		elsif ($info =~ /(\d+\.\d+)/) {
+			$version = $1;
+		}
+		else {
+			die "No version from initdb?! ($info)\n";
+		}
 
 		## Make sure pg_ctl is available as well before we go further
 		if (! -e $pg_ctl) {
@@ -268,10 +291,10 @@ sub connect_database {
 		}
 		$info = '';
 		eval {
-			$info = qx{$pg_ctl --help};
+			$info = qx{$pg_ctl --help 2>&1};
 		};
 		last GETHANDLE if $@; ## Fail - pg_ctl bad
-		if (!defined $info or $info !~ /\@postgresql\.org/) {
+		if (!defined $info or ($info !~ /\@postgresql\.org/ and $info !~ /run as root/)) {
 			$@ = defined $initdb ? "Bad pg_ctl output: $info" : 'Bad pg_ctl output';
 			last GETHANDLE; ## Fail - pg_ctl bad
 		}
@@ -300,7 +323,7 @@ sub connect_database {
 			if (open $fh, '>', $readme) {
 				print $fh "This is a test directory for DBD::Pg and may be removed\n";
 				print $fh "You may want to ensure the postmaster has been stopped first.\n";
-				print $fh "Check the port in the postgresql.conf file\n";
+				print $fh "Check the data/postmaster.pid file\n";
 				close $fh or die qq{Could not close "$readme": $!\n};
 			}
 
@@ -312,7 +335,8 @@ sub connect_database {
 			unshift @userlist, $username if defined $username and $username ne getpwent;
 
 			my %doneuser;
-			for $testuser (@userlist) {
+			for (@userlist) {
+				$testuser = $_;
 				next if $doneuser{$testuser}++;
 				$uid = (getpwnam $testuser)[2];
 				next if !defined $uid;
@@ -322,7 +346,7 @@ sub connect_database {
 				$su = $testuser;
 				$founduser++;
 				$info = '';
-				my $olddir = getcwd;
+				$olddir = getcwd;
 				eval {
 					chdir $testdir;
 					$info = qx{su -m $testuser -c "$initdb --locale=C -E UTF8 -D $testdir/data 2>&1"};
@@ -393,8 +417,25 @@ sub connect_database {
 			$@ = qq{Could not open "$conf": $!};
 			last GETHANDLE; ## Fail - no conf file
 		}
-		print $cfh "\n\n## DBD::Pg testing parameters\nport=$testport\nmax_connections=4\n";
-		print $cfh "listen_addresses='localhost'\n" if $^O =~ /Win32/;
+		print $cfh "\n\n## DBD::Pg testing parameters\n";
+		print $cfh "port=$testport\n";
+		print $cfh "max_connections=4\n";
+		if ($version >= 8.0) {
+			print $cfh "log_statement = 'all'\n";
+			print $cfh "log_line_prefix = '%m [%p] '\n";
+		}
+		else {
+			print $cfh "silent_mode = true\n";
+		}
+		if ($version == 8.1) {
+			print {$cfh} "redirect_stderr = on\n";
+		}
+
+		if ($version >= 8.3) {
+			print {$cfh} "logging_collector = on\n";
+		}
+		print $cfh "log_min_messages = 'DEBUG1'\n";
+		print $cfh "listen_addresses='127.0.0.1'\n" if $^O =~ /Win32/;
 		print $cfh "\n";
 		close $cfh or die qq{Could not close "$conf": $!\n};
 
@@ -416,9 +457,12 @@ sub connect_database {
 					}
 				}
 				$option = q{-o '-k socket'};
+				if ($version <= 8.0) {
+					$option = q{-o '-k dbdpg_test_database/data/socket'};
+				}
 			}
 			my $COM = qq{$pg_ctl $option -l $testdir/dbdpg_test.logfile -D $testdir/data start};
-			my $olddir = getcwd;
+		    $olddir = getcwd;
 			if ($su) {
 				chdir $testdir;
 				$COM = qq{su -m $su -c "$COM"};
@@ -443,13 +487,31 @@ sub connect_database {
 		else {
 			$testdsn .= ";host=$testdir/data/socket";
 		}
+
 		my $loop = 1;
 	  STARTUP: {
 			eval {
 				$dbh = DBI->connect($testdsn, $testuser, '',
 									{RaiseError => 1, PrintError => 0, AutoCommit => 1});
 			};
-			if ($@ =~ /starting up/ or $@ =~ /PGSQL\.$testport/) {
+			## Regardless of the error, try again.
+			## We used to check the message, but LANG problems may complicate that.
+
+			if ($@) {
+				if ($@ =~ /database "postgres" does not exist/) {
+					## Old server, so let's create a postgres database manually
+					sleep 2;
+					(my $tempdsn = $testdsn) =~ s/postgres/template1/;
+					eval {
+						$dbh = DBI->connect($tempdsn, $testuser, '',
+											{RaiseError => 1, PrintError => 0, AutoCommit => 1});
+					};
+					$dbh->do('CREATE DATABASE postgres');
+					$dbh->disconnect();
+					if ($@) {
+						die "Could not connect: $@\n";
+					}
+				}
 				if ($loop++ < 5) {
 					sleep 1;
 					redo STARTUP;
@@ -471,6 +533,7 @@ sub connect_database {
 		print $fh "## Helpconnect: $helpconnect\n";
 		print $fh "## pg_ctl: $pg_ctl\n";
 		print $fh "## initdb: $initdb\n";
+		print $fh "## Version: $version\n";
 		if ($connerror) {
 			print $fh "## DSN: FAIL!\n";
 			print $fh "## ERROR: $connerror\n";
@@ -492,6 +555,16 @@ sub connect_database {
 	$ENV{DBI_DSN} = $testdsn;
 	$ENV{DBI_USER} = $testuser;
 
+	if ($arg->{quickreturn}) {
+		return $helpconnect, '', $dbh;
+	}
+
+	my $SQL = 'SELECT usesuper FROM pg_user WHERE usename = current_user';
+	my $bga = $dbh->selectall_arrayref($SQL)->[0][0];
+	if ($bga) {
+		$dbh->do(q{SET LC_MESSAGES = 'C'});
+	}
+
 	if ($arg->{nosetup}) {
 		return $helpconnect, '', $dbh unless schema_exists($dbh, $S);
 		$dbh->do("SET search_path TO $S");
@@ -511,7 +584,7 @@ sub connect_database {
 		$dbh->do("SET search_path TO $S");
 		$dbh->do('CREATE SEQUENCE dbd_pg_testsequence');
 		# If you add columns to this, please do not use reserved words!
-		my $SQL = q{
+		$SQL = q{
 CREATE TABLE dbd_pg_test (
   id         integer not null primary key,
   lii        integer unique not null default nextval('dbd_pg_testsequence'),
@@ -522,7 +595,9 @@ CREATE TABLE dbd_pg_test (
   pdate      timestamp default now(),
   testarray  text[][],
   testarray2 int[],
+  testarray3 bool[],
   "CaseTest" boolean,
+  expo       numeric(6,2),
   bytetest   bytea
 )
 };
@@ -568,7 +643,7 @@ sub find_tempdir {
 
 sub get_test_settings {
 
-	## Returns test databae information from the testfile if it exists
+	## Returns test database information from the testfile if it exists
 	## Defaults to ENV variables or blank
 
 	## Find the best candidate for the pg_ctl program
@@ -577,7 +652,7 @@ sub get_test_settings {
 		($pg_ctl = $ENV{PGINITDB}) =~ s/initdb/pg_ctl/;
 	}
 	my ($testdsn, $testuser, $testdir, $error) = ('','','','?');
-	my ($helpconnect, $su, $uid, $initdb) = (0,'','','default');
+	my ($helpconnect, $su, $uid, $initdb, $version) = (0,'','','default',0);
 	my $inerror = 0;
 	if (-e $helpfile) {
 		open $fh, '<', $helpfile or die qq{Could not open "$helpfile": $!\n};
@@ -586,7 +661,7 @@ sub get_test_settings {
 				$error .= "\n$_";
 			}
 			/DSN: (.+)/           and $testdsn = $1;
-			/User: (\w+)/         and $testuser = $1;
+			/User: (\S+)/         and $testuser = $1;
 			/Helpconnect: (\d+)/  and $helpconnect = $1;
 			/Testowner: (\w+)/    and $su = $1;
 			/Testowneruid: (\d+)/ and $uid = $1;
@@ -594,6 +669,7 @@ sub get_test_settings {
 			/pg_ctl: (.+)/        and $pg_ctl = $1;
 			/initdb: (.+)/        and $initdb = $1;
 			/ERROR: (.+)/         and $error = $1 and $inerror = 1;
+			/Version: (.+)/       and $version = $1;
 		}
 		close $fh or die qq{Could not close "$helpfile": $!\n};
 	}
@@ -603,7 +679,7 @@ sub get_test_settings {
 		$testdir = "$dir/dbdpg_test_database";
 	}
 
-	return $testdsn, $testuser, $helpconnect, $su, $uid, $testdir, $pg_ctl, $initdb, $error;
+	return $testdsn, $testuser, $helpconnect, $su, $uid, $testdir, $pg_ctl, $initdb, $error, $version;
 
 } ## end of get_test_settings
 
@@ -674,7 +750,7 @@ sub shutdown_test_database {
 	my ($testdsn,$testuser,$helpconnect,$su,$uid,$testdir,$pg_ctl,$initdb) = get_test_settings();
 
 	if (-e $testdir and -e "$testdir/data/postmaster.pid") {
-		my $COM = qq{$pg_ctl -D $testdir/data --silent -m fast stop};
+		my $COM = qq{$pg_ctl -D $testdir/data -m fast stop};
 		my $olddir = getcwd;
 		if ($su) {
 			$COM = qq{su $su -m -c "$COM"};
